#include "oicosahedron.h"

//////////////////////////////////////////////////////////////////////
// OIcosahedron Class
//

OIcosahedron::OIcosahedron()
{
  m_fDihedralAngle = acos(-sqrt(5.0)/3.0); // 138.19
  //GoldenRatio = 0.5*(1+sqrt(5.0));
}

OIcosahedron::~OIcosahedron()
{
}

VOID OIcosahedron::InitFaceGrid()
{
  Vector v1, v2, v3;

  v1.Set( 0.0, GOLDENRATIO,         1.0);
  v2.Set(-1.0,         0.0, GOLDENRATIO);
  v3.Set( 1.0,         0.0, GOLDENRATIO);

  v1.Normalize();
  v2.Normalize();
  v3.Normalize();

  // 2D
  m_pTile->InsertFaceGridPoint( v1[X], v1[Y], v1[Z]);
  m_pTile->InsertFaceGridPoint( v2[X], v2[Y], v2[Z]);
  m_pTile->InsertFaceGridPoint( v3[X], v3[Y], v3[Z]);
}

VOID OIcosahedron::InitSphereFaceGrid()
{
  Point pt, pt1, pt2, pt3;

  // 3D
  pt1.Set( 0.0, GOLDENRATIO,         1.0);
  pt2.Set(-1.0,         0.0, GOLDENRATIO);
  pt3.Set( 1.0,         0.0, GOLDENRATIO);

  // [pt1 -> pt2 -> pt3]
  AddSphereFaceGridTesselatedPoints(pt1, pt2);
  AddSphereFaceGridTesselatedPoints(pt2, pt3);
  AddSphereFaceGridTesselatedPoints(pt3, pt1);
}

VOID OIcosahedron::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt1);
  tpt2->SetPartner(tpt4);
  tpt3->SetPartner(tpt3);
  tpt4->SetPartner(tpt2);

  m_pTile->TesselateOutlineAll();

  if( (GetType() == estOITile1) ||
      (GetType() == estOITile2) )
    {
      tpt2->SetHeightPartner(tpt4);
      tpt4->SetHeightPartner(tpt2);
    }
  else if( GetType() == estOITile3 )
    {
      tpt1->SetHeightPartner(tpt3);
      tpt3->SetHeightPartner(tpt1);
    }
  else
    {
      ASSERT(FALSE);
    }
}

VOID OIcosahedron::InitFaceMatrices()
{
  MATRIX4D m;   
  Matrix *mm;
  
  m_listFaceTransformMatrix.RemoveAll();
  
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();
  {
    glLoadIdentity();

    /////////////////////////////////////////////////////////////////////////
    // Top

    // Face 1
    glPushMatrix();
    {
      // Identity
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 2
    glPushMatrix();
    {
      glRotatef(1*72, 0.0, GOLDENRATIO, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 3
    glPushMatrix();
    {
      glRotatef(2*72, 0.0, GOLDENRATIO, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 4
    glPushMatrix();
    {
      glRotatef(3*72, 0.0, GOLDENRATIO, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 5
    glPushMatrix();
    {
      glRotatef(4*72, 0.0, GOLDENRATIO, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Top Side

    // Face 6
    glPushMatrix();
    {
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 7
    glPushMatrix();
    {
      glRotatef(1*72, 0.0, GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 8
    glPushMatrix();
    {
      glRotatef(2*72, 0.0, GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 9
    glPushMatrix();
    {
      glRotatef(3*72, 0.0, GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 10
    glPushMatrix();
    {
      glRotatef(4*72, 0.0, GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Bottom

    // Face 11
    glPushMatrix();
    {
      glRotatef(2*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 12
    glPushMatrix();
    {
      glRotatef(1*72, 0.0, -GOLDENRATIO, -1.0);
      glRotatef(2*72, 0.0, -GOLDENRATIO,  1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 13
    glPushMatrix();
    {
      glRotatef(2*72, 0.0, -GOLDENRATIO, -1.0);
      glRotatef(2*72, 0.0, -GOLDENRATIO,  1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 14
    glPushMatrix();
    {
      glRotatef(3*72, 0.0, -GOLDENRATIO, -1.0);
      glRotatef(2*72, 0.0, -GOLDENRATIO,  1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 15
    glPushMatrix();
    {
      glRotatef(4*72, 0.0, -GOLDENRATIO, -1.0);
      glRotatef(2*72, 0.0, -GOLDENRATIO,  1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    /////////////////////////////////////////////////////////////////////////
    // Bottom Side

    // Face 16
    glPushMatrix();
    {
      glRotatef(1*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 17
    glPushMatrix();
    {
      glRotatef(1*72, 0.0,  GOLDENRATIO, 1.0);
      glRotatef(1*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 18
    glPushMatrix();
    {
      glRotatef(2*72, 0.0,  GOLDENRATIO, 1.0);
      glRotatef(1*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 19
    glPushMatrix();
    {
      glRotatef(3*72, 0.0,  GOLDENRATIO, 1.0);
      glRotatef(1*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();

    // Face 20
    glPushMatrix();
    {
      glRotatef(4*72, 0.0,  GOLDENRATIO, 1.0);
      glRotatef(1*72, 0.0, -GOLDENRATIO, 1.0);
      glScalef(-1.0, -1.0, 1.0);
      glGetFloatv(GL_MODELVIEW_MATRIX, m);
      mm = new Matrix();
      mm->Set(m);
      m_listFaceTransformMatrix.Insert(mm, 0);
    }
    glPopMatrix();
  }
  glPopMatrix();
}

VOID OIcosahedron::InitTileMatrices()
{
  Matrix *mm;
  Matrix m1, m2;
  Vector v;

  m_listTileTransformMatrix.RemoveAll();

  v.Set( 0.0, GOLDENRATIO/3.0, (2*GOLDENRATIO + 1.0)/3.0 );

  // Tile 1
  mm = new Matrix(); // Identity
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 2
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(120));
  m_listTileTransformMatrix.Insert(mm, 0);

  // Tile 3
  mm = new Matrix();
  mm->SetRotate(v, DEGTORAD(240));
  m_listTileTransformMatrix.Insert(mm, 0);
}

VOID OIcosahedron::FindMatrices(Matrix &/*m1*/, Matrix &/*m2*/, TilePoint */*tpt1*/)
{
  ASSERT(FALSE);
}

///////////////////////////////////////////////////////////////////////////////

VOID OIcosahedron::OutputES(FILE *fpOutputFile, CHAR *paletteFilename,
			    CHAR *colorFilename)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "icosahedron\n\n" );

  m_pTile->OutputES(fpOutputFile);

  fprintf(fpOutputFile, "palettefile %s\n", paletteFilename);
  fprintf(fpOutputFile, "colorindexfile %s\n", colorFilename);
  //OutputESColor(fpOutputFile);
}
