#include "symmetrygroup.h"
#include "tile.h"


enum GridType 
{
  gtNO,
  gtFACE,
  gtTILE,
  gtSize
};

extern GLuint GLuiSphere1_30;
extern GLuint GLuiSphere02_8;
extern GridType gtDrawGrid;
extern BOOL gbDrawNorthPole;

extern UINT_32 guiSpecialUI;

//////////////////////////////////////////////////////////////////////
// SymmetryGroup Class
//

SymmetryGroup::SymmetryGroup()
{
  GLuiSolid = glGenLists(1);
  GLuiBasRelief = glGenLists(1);

  m_bMakeSolidDisplayList = TRUE;
  m_bMakeBasReliefDisplayList = TRUE;

  m_bAllSolidTiles = TRUE;
  m_bTriangulated = FALSE;

  m_pTile = new Tile();
  m_pTile->SetSGParent(this);

  m_listFaceTransformMatrix.Init();
  m_listTileTransformMatrix.Init();

  //WM
  m_uiTiles = 0;
  m_puiTileColorIndex = NULL;
  m_pvTileColor = NULL;  
}

SymmetryGroup::~SymmetryGroup()
{
}

VOID SymmetryGroup::Init()
{
  InitFaceGrid();
  InitTileGrid();

  InitSphereFaceGrid();
  InitSphereTileGrid();
  InitOutline();
  
  InitFaceMatrices();
  InitTileMatrices();

  UINT_32 i, uiFaces, uiTilesPerFace;
  uiFaces = m_listFaceTransformMatrix.GetLength();
  uiTilesPerFace = m_listTileTransformMatrix.GetLength();

  //WM
  m_uiTiles = uiFaces * uiTilesPerFace;
  m_puiTileColorIndex = new UINT_32[m_uiTiles];
  m_pvTileColor = new Vector[60];

  for (i = 0; i < m_uiTiles; i++)
      m_puiTileColorIndex[i] = 99;
  for (i = 0; i < 60; i++)
      m_pvTileColor[i].Set(0.5, 0.5, 0.5);
}

VOID SymmetryGroup::InitTileGrid()
{
  // 2D 
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint1 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint2 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint3 );
  m_pTile->InsertTileGridPoint( m_pt2DTilePoint4 );
}

VOID SymmetryGroup::InitSphereTileGrid()
{
  // [pt1 -> pt2 -> pt3 -> pt4]
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint1, m_pt2DTilePoint2);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint2, m_pt2DTilePoint3);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint3, m_pt2DTilePoint4);
  AddSphereTileGridTesselatedPoints(m_pt2DTilePoint4, m_pt2DTilePoint1);
}

///////////////////////////////////////////////////////////////////////////////

VOID SymmetryGroup::AddSphereTileGridTesselatedPoints(Point pt1, Point pt2)
{
  UINT_32 i;
  FLOAT fSlices;
  Point pt;
  Vector v;

  fSlices = 50.0;

  // [pt1 -> pt2)
  for( i = 0; i < fSlices; i++ )
    {
      pt = ((FLOAT)i/fSlices)*pt2 + (1.0 - ((FLOAT)i/fSlices))*pt1;
      v = pt;
      v[W] = 0.0;
      v.Normalize();
      pt = v;
      pt = 1.0025 * pt;
      m_pTile->InsertSphereTileGridPoint(pt);
    }
}

VOID SymmetryGroup::AddSphereFaceGridTesselatedPoints(Point pt1, Point pt2)
{
  UINT_32 i;
  FLOAT fSlices;
  Point pt;
  Vector v;

  fSlices = 50.0;

  // [pt1 -> pt2)
  for( i = 0; i < fSlices; i++ )
    {
      pt = ((FLOAT)i/fSlices)*pt2 + (1.0 - ((FLOAT)i/fSlices))*pt1;
      v = pt;
      v[W] = 0.0;
      v.Normalize();
      pt = v;
      pt = 1.0025 * pt;
      m_pTile->InsertSphereFaceGridPoint(pt);
    }
}

VOID SymmetryGroup::MoveOutlinePoint(UINT_32 uiSelectedPoint, Point pt)
{
  m_bTriangulated = FALSE;
  m_pTile->MoveOutlinePoint(uiSelectedPoint, pt);
}

VOID SymmetryGroup::InsertOutlinePoint(UINT_32 uiSelectedPoint, Point ptLocation)
{
  m_bTriangulated = FALSE;
  m_pTile->InsertOutlinePoint(uiSelectedPoint, ptLocation);
}

VOID SymmetryGroup::DeleteOutlinePoint(UINT_32 uiSelectedPoint)
{
  m_bTriangulated = FALSE;
  m_pTile->DeleteOutlinePoint(uiSelectedPoint);
}

TilePoint *SymmetryGroup::AddInnerPoint(Point pt)
{
  m_bTriangulated = FALSE;
  return (m_pTile->AddInnerPoint(pt));
}

TilePoint *SymmetryGroup::AddInnerPointCopy(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  m_bTriangulated = FALSE;
  return (m_pTile->AddInnerPointCopy(uiSelectedType, uiSelectedPoint));
}

VOID SymmetryGroup::MoveInnerPoint(UINT_32 uiSelectedPoint, Point pt)
{
  m_bTriangulated = FALSE;
  m_pTile->MoveInnerPoint(uiSelectedPoint, pt);
}

VOID SymmetryGroup::InsertInnerPoint(UINT_32 uiSelectedPoint, Point pt)
{
  m_bTriangulated = FALSE;
  m_pTile->InsertInnerPoint(uiSelectedPoint, pt);
}

VOID SymmetryGroup::DeleteInnerPoint(UINT_32 uiSelectedPoint)
{
  m_bTriangulated = FALSE;
  m_pTile->DeleteInnerPoint(uiSelectedPoint);
}

VOID SymmetryGroup::SnapInnerPointToSelected(UINT_32 uiInnerPointIndex, 
					     UINT_32 uiSelectedType,
					     UINT_32 uiSelectedIndex)
{
  m_bTriangulated = FALSE;
  m_pTile->SnapInnerPointToSelected(uiInnerPointIndex, uiSelectedType, uiSelectedIndex);
}

VOID SymmetryGroup::ClearSnapPoint(UINT_32 uiInnerPointIndex)
{
  m_pTile->ClearSnapPoint(uiInnerPointIndex);
}

VOID SymmetryGroup::AddHolePoint(Point pt)
{
  m_bTriangulated = FALSE;
  m_pTile->AddHolePoint(pt);
}

VOID SymmetryGroup::MoveHolePoint(UINT_32 uiSelectedPoint, Point pt)
{
  m_bTriangulated = FALSE;
  m_pTile->MoveHolePoint(uiSelectedPoint, pt);
}

VOID SymmetryGroup::DeleteHolePoint(UINT_32 uiSelectedPoint)
{
  m_bTriangulated = FALSE;
  m_pTile->DeleteHolePoint(uiSelectedPoint);
}

BOOL SymmetryGroup::GetHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  return (m_pTile->GetPointHeightSelected(uiSelectedType, uiSelectedPoint));
}

VOID SymmetryGroup::SetHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint, 
				      BOOL b)
{
  m_pTile->SetPointHeightSelected(uiSelectedType, uiSelectedPoint, b);
}

VOID SymmetryGroup::TogleHeightSelected(UINT_32 uiSelectedType, UINT_32 uiSelectedPoint)
{
  m_pTile->ToglePointHeightSelected(uiSelectedType, uiSelectedPoint);
}

Vector SymmetryGroup::GetPointNormalVector(UINT_32 uiSelectedPoint)
{
  return (m_pTile->GetPointNormalVector(uiSelectedPoint));
}

VOID SymmetryGroup::AddHeightToPoints(FLOAT fHeight)
{
  m_pTile->AddHeightToPoints(fHeight);
}

VOID SymmetryGroup::DeselectAllHeightPoints()
{
  m_pTile->DeselectAllHeightPoints();
}

//////////////////////////////////////////////////////////////////////////////////

VOID SymmetryGroup::CalculateStereographicPoints()
{
  m_pTile->CalculateStereographicPoints();
}

VOID SymmetryGroup::TriangulateTile()
{
  if( !m_bTriangulated )
    {
      m_pTile->Triangulate();
      m_bTriangulated = TRUE;

      m_pTile->SetBasReliefModified(TRUE);
    }
}

//WM
VOID SymmetryGroup::SetIndexTile(UINT_32 uiSelectedTile, UINT_32 indexColor) 
{
  if (uiSelectedTile < m_uiTiles)
    m_puiTileColorIndex[uiSelectedTile] = indexColor;
  return;
}

//WM
VOID SymmetryGroup::SetColorIndex(UINT_32 uiSelectedIndex, Vector vColor)
{
  ASSERT(uiSelectedIndex >= 0 && uiSelectedIndex <=59);
  m_pvTileColor[uiSelectedIndex].Set(vColor.Get());
  return;
}

//WM 
VOID SymmetryGroup::PaintTile(UINT_32 uiSelectedTile, UINT_32 uiSelectedIndex)
{
  ASSERT(uiSelectedIndex >= 0 && uiSelectedIndex <=59);
  m_puiTileColorIndex[uiSelectedTile] = uiSelectedIndex;
  //m_pvTileColor[uiSelectedIndex].Set(vColor.Get());
  return;
}

//WM
UINT_32 SymmetryGroup::GetColorIndex(UINT_32 uiSelectedTile) 
{
  return m_puiTileColorIndex[uiSelectedTile];
}

VOID SymmetryGroup::SetRadius(FLOAT fRadius)
{
  m_pTile->SetRadius(fRadius);
}

VOID SymmetryGroup::SetThickness(FLOAT fThickness)
{
  m_pTile->SetThickness(fThickness);
}

//WM
VOID SymmetryGroup::OutputCES(FILE *fpOutputFile, CHAR *symmetryGroupType)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 i;

  fprintf(fpOutputFile, "%s\n\n", symmetryGroupType);

  fprintf(fpOutputFile, "tiles %lu\n", m_uiTiles);

  for (i = 0; i < m_uiTiles; i++) 
    fprintf(fpOutputFile, "tile%lu %ld\n", i, m_puiTileColorIndex[i]); 
  fprintf(fpOutputFile, "\n");
}

//WM
VOID SymmetryGroup::OutputPES(FILE *fpOutputFile)
{
  ASSERT ( fpOutputFile != NULL);

  UINT_32 i, length;
  Vector v;

  length = 60;
  fprintf(fpOutputFile, "colors %lu\n", length);
  for (i = 0; i < length; i++)
    {
      v = m_pvTileColor[i];
      fprintf(fpOutputFile, "rgb%lu %.6f %.6f %.6f\n", i, v[X], v[Y], v[Z]);
    }
  fprintf(fpOutputFile, "\n");
}

VOID SymmetryGroup::OutputSTL(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  CListIterInt iter, iter2;
  Matrix *mm1, *mm2;
  Matrix m, m1, m2;

  fprintf(fpOutputFile, "solid\n" );

  if( !m_bAllSolidTiles )
    {
      m_pTile->OutputSTL(fpOutputFile, m, TRUE);
    }
  else
    {
      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( mm1 = (Matrix *)(iter.PeekFirst()); 
	   mm1 != NULL;
	   mm1 = (Matrix *)(iter.PeekNext()) )
	{
	  m1.Set(mm1->Get());
	  
	  // For each tile apply the appropriate transform
	  iter2.Init(&m_listTileTransformMatrix);
	  for( mm2 = (Matrix *)(iter2.PeekFirst()); 
	       mm2 != NULL;
	       mm2 = (Matrix *)(iter2.PeekNext()) )
	    {
	      m2.Set(mm2->Get());
	      
	      m = m1 * m2;
	      
	      m_pTile->OutputSTL(fpOutputFile, m, FALSE);
	    }
	}
    }

  fprintf(fpOutputFile, "endsolid\n" );
}

VOID SymmetryGroup::OutputVRML(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  UINT_32 i, j;
  CListIterInt iter, iter2, iter3;
  Matrix *mm1, *mm2;
  Matrix m, m1, m2;
  Vector greyColor;
  Vector *vColor;
  Vector v;

  UINT_32 k, idColor;

  greyColor.Set(0.5, 0.5, 0.5);

  fprintf(fpOutputFile, "#VRML V2.0 utf8\n\n");
  fprintf(fpOutputFile, "Background {\n");
  fprintf(fpOutputFile, "  skyColor [0.0 0.0 0.0]\n");
  fprintf(fpOutputFile, "}\n\n");

  k = 0;
  idColor = m_puiTileColorIndex[k];
  if(idColor == 99)
    vColor = &greyColor;
  else
    {
      ASSERT(idColor >= 0 && idColor <= 59);
      vColor = &(m_pvTileColor[idColor]);
    }

  if( !m_bAllSolidTiles )
    {
      ASSERT( vColor != NULL );
      v.Set(vColor->Get());

      m_pTile->EnumerateOutlinePoints();

      m_pTile->OutputVRML(fpOutputFile, m, v, TRUE);
    }
  else
    {
      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( mm1 = (Matrix *)(iter.PeekFirst()), i = 0; 
	   mm1 != NULL;
	   mm1 = (Matrix *)(iter.PeekNext()), i++ )
	{
	  m1.Set(mm1->Get());
	  
	  // For each tile apply the appropriate transform
	  iter2.Init(&m_listTileTransformMatrix);
	  for( mm2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
	       mm2 != NULL;
	       mm2 = (Matrix *)(iter2.PeekNext()), j++ )
	    {
	      m2.Set(mm2->Get());
	      
	      m = m1 * m2;
	      
	      ASSERT( vColor != NULL );
	      v.Set(vColor->Get());

	      m_pTile->OutputVRML(fpOutputFile, m, v, FALSE);
	      
	      k++;
	      if (k < m_uiTiles)
		idColor = m_puiTileColorIndex[k];
	      if(idColor == 99)
		vColor = &greyColor;
	      else
		{
		  ASSERT(idColor >= 0 && idColor <= 59);
		  vColor = &(m_pvTileColor[idColor]);
		}
	    }
	}
    }

  fprintf(fpOutputFile, "NavigationInfo {\n");
  fprintf(fpOutputFile, "  type \"EXAMINE\"\n");
  fprintf(fpOutputFile, "}\n\n");
}

VOID SymmetryGroup::ClearSphereOutlinePoints()
{
  m_pTile->ClearSphereOutlinePoints();
}

VOID SymmetryGroup::ClearSphereInnerPoints()
{
  m_pTile->ClearSphereInnerPoints();
}

VOID SymmetryGroup::ClearSphereHolePoints()
{
  m_pTile->ClearSphereHolePoints();
}

VOID SymmetryGroup::CreateTilePoint(Point pt, UINT_32 uiType, UINT_32 uiTilePoint, 
				    UINT_32 uiConnected, FLOAT fHeight, Matrix m)
{
  m_pTile->CreateTilePoint(pt, uiType, uiTilePoint, uiConnected, fHeight, m);
}

VOID SymmetryGroup::SetTilePointPartners(UINT_32 uiID, UINT_32 uiIDPartner)
{
  m_pTile->SetPartners(uiID, uiIDPartner);
}

VOID SymmetryGroup::SetTileHeightPartnersFromOpenES(UINT_32 uiID, UINT_32 uiIDPartner)
{
  m_pTile->SetHeightPartnersFromOpenES(uiID, uiIDPartner);
}

VOID SymmetryGroup::TesselateTileOutlinePoints()
{
  m_pTile->TesselateOutlineAll();
}

VOID SymmetryGroup::TesselateTileInnerPoints()
{
  m_pTile->TesselateInnerAll();
}

///////////////////////////////////////////////////////////////////////////////

VOID SymmetryGroup::RenderSelectionSphereTileOutlinePoints()
{
  m_pTile->RenderSelectionSphereOutlinePoints();
}

VOID SymmetryGroup::RenderSelectionSphereTileOutlineEdges()
{
  m_pTile->RenderSelectionSphereOutlineEdges();
}

VOID SymmetryGroup::RenderSelectionSphereTileInnerPoints()
{
  m_pTile->RenderSelectionSphereInnerPoints();
}

VOID SymmetryGroup::RenderSelectionSphereTileInnerEdges()
{
  m_pTile->RenderSelectionSphereInnerEdges();
}

VOID SymmetryGroup::RenderSelectionSphereTileHeightOutlinePoints()
{
  m_pTile->RenderSelectionSphereHeightOutlinePoints();
}

VOID SymmetryGroup::RenderSelectionSphereTileHeightInnerPoints()
{
  m_pTile->RenderSelectionSphereHeightInnerPoints();
}

VOID SymmetryGroup::RenderSelectionSphereTileHolePoints()
{
  m_pTile->RenderSelectionSphereHolePoints();
}

VOID SymmetryGroup::RenderPolyhedra()
{
  CListIterInt iter, iter2;
  Matrix *m, *m2;

  glMatrixMode(GL_MODELVIEW);

  glPushAttrib(GL_LINE_BIT | GL_COLOR_BUFFER_BIT);
  {
    glLineWidth(2.0);

    // For each face apply the appropriate transform
    iter.Init(&m_listFaceTransformMatrix);
    for( m = (Matrix *)(iter.PeekFirst()); 
	 m != NULL;
	 m = (Matrix *)(iter.PeekNext()) )
      {
	glPushMatrix();
	{
	  glMultMatrixf(m->Get());
	  
	  glDisable(GL_LIGHTING);
	  {	    
	    // Red Face Outline
	    if( gtDrawGrid == gtFACE )
	      {
		glColor3f(1.0, 0.0, 0.0);
		glBegin(GL_LINE_LOOP);
		{
		  m_pTile->PrintVertexFacePoints();
		}
		glEnd();
	      }

	    // Yellow Tiles
	    if( gtDrawGrid == gtTILE )
	      {
		glColor3f(1.0, 1.0, 0.0);	    

		// For each tile apply the appropriate transform
		iter2.Init(&m_listTileTransformMatrix);
		for( m2 = (Matrix *)(iter2.PeekFirst()); 
		     m2 != NULL;
		     m2 = (Matrix *)(iter2.PeekNext()) )
		  {
		    glPushMatrix();
		    {
		      glMultMatrixf(m2->Get());
		      glBegin(GL_LINE_LOOP);
		      {
			m_pTile->PrintVertexTilePoints();
		      }
		      glEnd();
		    }
		    glPopMatrix();
		  }
	      }
	  }
	  glEnable(GL_LIGHTING);

	  // Grey Face
	  glColor3f(0.5, 0.5, 0.5);
	  glBegin(GL_POLYGON);
	  {
	    m_pTile->PrintVertexFacePoints();
	  }
	  glEnd();	  
	}
	glPopMatrix();
      }
  }
  glPopAttrib();
}

VOID SymmetryGroup::RenderBall()
{
  UINT_32 i, j;
  CListIterInt iter, iter2;
  Matrix *m, *m2;

  glMatrixMode(GL_MODELVIEW);

  // Sphere
  glColor3f(0.5, 0.5, 0.5);	
  glCallList(GLuiSphere1_30);

  // Edges
  glDisable(GL_LIGHTING);
  {
    glPushAttrib(GL_LINE_BIT | GL_COLOR_BUFFER_BIT);
    {
      glLineWidth(2.0);

      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( m = (Matrix *)(iter.PeekFirst()), i = 0; 
	   m != NULL;
	   m = (Matrix *)(iter.PeekNext()), i++ )
	{
	  glPushMatrix();
	  {
	    glMultMatrixf(m->Get());

	    if( gtDrawGrid == gtFACE )
	      {
		glColor3f(1.0, 0.0, 0.0);
		m_pTile->RenderSphereFace();
	      }

	    // For each tile apply the appropriate transform
	    iter2.Init(&m_listTileTransformMatrix);
	    for( m2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
		 m2 != NULL;
		 m2 = (Matrix *)(iter2.PeekNext()), j++ )
	      {
		glPushMatrix();
		{
		  glMultMatrixf(m2->Get());

		  if( gtDrawGrid == gtTILE )
		    {
		      glColor3f(1.0, 1.0, 0.0);
		      m_pTile->RenderSphereTile();
		    }

		  glColor3f(1.0, 1.0, 1.0); // white
		  m_pTile->RenderSphereOutlineEdges();
    
		  if( (i == 0) && (j == 0) )
		    glColor3f(1.0, 0.0, 1.0); // magenta

		  m_pTile->RenderSphereInnerEdges();
		}
		glPopMatrix();
	      }
	  }
	  glPopMatrix();
	}

      // Draw the tile outline again so it stands out
      glLineWidth(3.0);
      glColor3f(0.0, 0.0, 1.0); // blue
      m_pTile->RenderSphereRaisedOutlineEdges();

    }
    glPopAttrib();
  }
  glEnable(GL_LIGHTING);

  // Points
  m_pTile->RenderSphereOutlinePoints();
  m_pTile->RenderSphereInnerPoints();
  m_pTile->RenderSphereHolePoints();

  // North Pole
  if( gbDrawNorthPole )
    {
      Matrix mm;
      Vector v;
      m_pTile->CalculateNorthPole();
      glColor3f(0.0, 1.0, 0.0);
      
      glPushMatrix();
      {  
	v = m_pTile->GetNorthPole();
	v[W] = 1.0;
	mm.SetTranslate(v);
	glMultMatrixf(mm.Get());
	glCallList(GLuiSphere02_8);
      }
      glPopMatrix();
      
      glPushMatrix();
      {  
	v = -v;
	v[W] = 1.0;
	mm.SetTranslate(v);
	glMultMatrixf(mm.Get());
	glCallList(GLuiSphere02_8);
      }
      glPopMatrix();
    }
}

VOID SymmetryGroup::RenderStereographicPoints()
{
  CListIterInt iter, iter2;

  glMatrixMode(GL_MODELVIEW);

  glPushAttrib(GL_POLYGON_BIT | GL_LINE_BIT);
  {
    glLineWidth(2.0);
    m_pTile->RenderStereographicPoints();

    glDisable(GL_LIGHTING);
    {
      glColor3f(0.0, 1.0, 0.0);
      m_pTile->RenderStereographicEdges();

      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      glColor3f(1.0, 1.0, 0.0);
      m_pTile->RenderMesh();
    }
    glEnable(GL_LIGHTING);
  }
  glPopAttrib();
}

VOID SymmetryGroup::RenderSphereTriangulation()
{
  if( m_bMakeBasReliefDisplayList || m_pTile->GetBasReliefModified() )
    {
      glDeleteLists( GLuiBasRelief, 1 );

      glNewList( GLuiBasRelief, GL_COMPILE );
      {
	// Render

	glColor3f(0.5, 0.5, 0.5);
	m_pTile->RenderMeshTilePoints();
	
	// Points
	glMatrixMode(GL_MODELVIEW);
	
	m_pTile->RenderHeightOutlinePoints();
	m_pTile->RenderHeightInnerPoints();
	
	// Edges
	glDisable(GL_LIGHTING);
	{
	  glPushAttrib(GL_LINE_BIT | GL_COLOR_BUFFER_BIT);
	  {
	    glLineWidth(2.0);
	    
	    // Outer
	    glColor3f(0.0, 0.0, 1.0); // blue
	    m_pTile->RenderHeightOutlineEdges();
	    
	    glColor3f(1.0, 1.0, 1.0); // white  
	    m_pTile->RenderHeightOutlinePointNormals();
	    
	    // Inner
	    glColor3f(1.0, 0.0, 1.0); // magenta	    
	    m_pTile->RenderHeightInnerEdges();
	    
	    glColor3f(1.0, 1.0, 1.0); // white  
	    m_pTile->RenderHeightInnerPointNormals();
	  }
	  glPopAttrib();
	}
	glEnable(GL_LIGHTING);
      }
      glEndList();
      
      m_pTile->SetBasReliefModified(FALSE);
      m_bMakeBasReliefDisplayList = FALSE;
    }

  glCallList(GLuiBasRelief);
}

//WM
VOID SymmetryGroup::RenderSolid()
{
  UINT_32 i, j;
  CListIterInt iter, iter2, iter3;
  Matrix *m, *m2;
  Vector greyColor;
  Vector *vColor;

  UINT_32 k, idColor;
      
  greyColor.Set(0.5, 0.5, 0.5);

  if( m_bMakeSolidDisplayList || m_pTile->GetSolidModified() )
    {
      glDeleteLists( GLuiSolid, 1 );

      glNewList( GLuiSolid, GL_COMPILE );
      {
	if( !m_bAllSolidTiles )
	  {
	    m_pTile->RenderMeshExtrudedTilePoints(TRUE);
	  }
	else
	  {
	    m_pTile->RenderMeshExtrudedTilePoints(FALSE);
	  }
      }
      glEndList();
      
      m_pTile->SetSolidModified(FALSE);
      m_bMakeSolidDisplayList = FALSE;
    }

  k = 0;
  idColor = m_puiTileColorIndex[k];
  if(idColor == 99)
    vColor = &greyColor;
  else
    {
      ASSERT(idColor >= 0 && idColor <= 59);
      vColor = &(m_pvTileColor[idColor]);
    }

  if( !m_bAllSolidTiles )
    {
      ASSERT( vColor != NULL );
      glColor3fv(vColor->Get());

      //m_pTile->RenderMeshExtrudedTilePoints(TRUE);
      glCallList(GLuiSolid);
    }
  else
    {
      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( m = (Matrix *)(iter.PeekFirst()), i = 0; 
	   m != NULL;
	   m = (Matrix *)(iter.PeekNext()), i++ )
	{
	  glPushMatrix();
	  {
	    glMultMatrixf(m->Get());
	    
	    // For each tile apply the appropriate transform
	    iter2.Init(&m_listTileTransformMatrix);
	    for( m2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
		 m2 != NULL;
		 m2 = (Matrix *)(iter2.PeekNext()), j++ )
	      {
		glPushMatrix();
		{
		  glMultMatrixf(m2->Get());

		  ASSERT( vColor != NULL );
		  glColor3fv(vColor->Get());
		  //m_pTile->RenderMeshExtrudedTilePoints(FALSE);
		  glCallList(GLuiSolid);

		  k++;
		  if(k < m_uiTiles)
		    idColor = m_puiTileColorIndex[k];
		  if(idColor == 99)
		    vColor = &(greyColor);
		  else 
		    {
		      ASSERT(idColor >= 0 && idColor <= 59);
		      vColor = &(m_pvTileColor[idColor]);
		    }
		}
		glPopMatrix();
	      }
	  }
	  glPopMatrix();
	}
    }


  // I3D Video
  
  /*
  if( !m_bMadeDisplayList )
    {
      glNewList( GLuiSolid, GL_COMPILE );
      {
	m_pTile->RenderMeshExtrudedTilePoints(TRUE);
      }
      glEndList();
      
      m_bMadeDisplayList = TRUE;
    }


  ASSERT( vColor != NULL );
  glColor3fv(vColor->Get());

  if( guiSpecialUI == 0 )
    {
      m_pTile->RenderMeshExtrudedTilePoints(TRUE);
    }
  else if( guiSpecialUI == 1 )
    {
      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( m = (Matrix *)(iter.PeekFirst()), i = 0; 
	   m != NULL;
	   m = (Matrix *)(iter.PeekNext()), i++ )
	{
	  glPushMatrix();
	  {
	    glMultMatrixf(m->Get());
	    
	    // For each tile apply the appropriate transform
	    iter2.Init(&m_listTileTransformMatrix);
	    for( m2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
		 m2 != NULL;
		 m2 = (Matrix *)(iter2.PeekNext()), j++ )
	      {
		if( ((i == 0) && (j == 0)) ||
		    ((i == 2) && (j == 2)) )
		  {
		    glPushMatrix();
		    {
		      glMultMatrixf(m2->Get());
		      
		      ASSERT( vColor != NULL );
		      glColor3fv(vColor->Get());
		      glCallList(GLuiSolid);
		      //m_pTile->RenderMeshExtrudedTilePoints(FALSE);
		    }
		    glPopMatrix();
		  }
		vColor = (Vector *)(iter3.PeekNext());
	      }
	  }
	  glPopMatrix();
	}
    }
  else
    {
      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( m = (Matrix *)(iter.PeekFirst()), i = 0; 
	   m != NULL;
	   m = (Matrix *)(iter.PeekNext()), i++ )
	{
	  glPushMatrix();
	  {
	    glMultMatrixf(m->Get());
	    
	    // For each tile apply the appropriate transform
	    iter2.Init(&m_listTileTransformMatrix);
	    for( m2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
		 m2 != NULL;
		 m2 = (Matrix *)(iter2.PeekNext()), j++ )
	      {
		glPushMatrix();
		{
		  glMultMatrixf(m2->Get());

		  ASSERT( vColor != NULL );
		  glColor3fv(vColor->Get());
		  glCallList(GLuiSolid);
		  //m_pTile->RenderMeshExtrudedTilePoints(FALSE);

		  vColor = (Vector *)(iter3.PeekNext());
		}
		glPopMatrix();
	      }
	  }
	  glPopMatrix();
	}
    }
    */
}

VOID SymmetryGroup::RenderSelectionSolid()
{
  if( !m_bAllSolidTiles )
    {
      glPushName(0);
      {
	m_pTile->RenderMeshExtrudedTilePoints(TRUE);
      }
    }
  else
    {
      UINT_32 i, j, uiName;
      CListIterInt iter, iter2;
      Matrix *m, *m2;
      
      uiName = 0;

      // For each face apply the appropriate transform
      iter.Init(&m_listFaceTransformMatrix);
      for( m = (Matrix *)(iter.PeekFirst()), i = 0; 
	   m != NULL;
	   m = (Matrix *)(iter.PeekNext()), i++ )
	{
	  glPushMatrix();
	  {
	    glMultMatrixf(m->Get());
	    
	    // For each tile apply the appropriate transform
	    iter2.Init(&m_listTileTransformMatrix);
	    for( m2 = (Matrix *)(iter2.PeekFirst()), j = 0; 
		 m2 != NULL;
		 m2 = (Matrix *)(iter2.PeekNext()), j++, uiName++ )
	      {
		glPushMatrix();
		{
		  glMultMatrixf(m2->Get());

		  glPushName(uiName);
		  {
		    m_pTile->RenderMeshExtrudedTilePoints(FALSE);
		  }
		  glPopName();
		}
		glPopMatrix();
	      }
	  }
	  glPopMatrix();
	}
    }
}

