#include "togl.h"
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>

#include "trackball.h"
#include "symmetrygroup.h"
// OTetrahedron
#include "ottile1.h"
#include "ottile2.h"
#include "stetrahedron.h"
// OOctahedron
#include "ootile1.h"
#include "ootile2.h"
#include "ootile3.h"
// OCube
#include "octile1.h"
#include "octile2.h"
#include "octile3.h"
// OIcosahedron
#include "oitile1.h"
#include "oitile2.h"
#include "oitile3.h"
// ODodecahedron
#include "odtile1.h"
#include "odtile2.h"
#include "odtile3.h"
// PentaDodecahedron
#include "pentadodecahedron.h"

#if defined(WIN32)
EXTERN int		TkConsoleInit(Tcl_Interp *interp);
#endif /* WIN32 */


// Called by Tk_Main() to let me init the modules (Togl) I will need.
int Tcl_AppInit(Tcl_Interp *pInterp);


// Selection types
#define SELECTED_NOTHING       1234567890
#define SELECTED_OUTLINE_POINT 1234567891
#define SELECTED_OUTLINE_EDGE  1234567892
#define SELECTED_INNER_POINT   1234567893
#define SELECTED_INNER_EDGE    1234567894

enum DrawType 
{
  dtSPHERE,
  dtPOLYHEDRA,
  dtSTEREO,
  dtTRI_TILE,
  dtSOLID,
  dtSize
};

enum GridType 
{
  gtNO,
  gtFACE,
  gtTILE,
  gtSize
};

enum SymmetryGroupType 
{
  sgtOTetrahedron,
  sgtSTetrahedron,
  sgtOOctahedron,
  sgtSOctahedron,
  sgtOCube,
  sgtSCube,
  sgtOIcosahedron,
  sgtSIcosahedron,
  sgtODodecahedron,
  sgtSDodecahedron,
  sgtPentaDodecahedron,
  sgtSize
};

static Togl *toglBall = NULL;
static Tcl_Interp *pTclInterp;
static BOOL bGUICreated = FALSE;
GLuint GLuiSphere1_30;
GLuint GLuiSphere01_10;
GLuint GLuiSphere02_10;


static char gpcBuffer[200];
static SymmetryGroup *pSG = NULL;
static TilePoint *tptLastInnerPoint = NULL;
static DrawType dtDrawType = dtSPHERE;
static BOOL bDrawInnerPoints = FALSE;
GridType gtDrawGrid = gtFACE;

// Selection
static UINT_32 uiSelectedType  = SELECTED_NOTHING;
static UINT_32 uiSelectedIndex = SELECTED_NOTHING;
static BOOL bLeftButtonDown = FALSE;
static BOOL bMiddleButtonDown = FALSE;
static BOOL bRightButtonDown = FALSE;
static BOOL bDeletePoint = FALSE;

// Ball Camera UI
static Matrix mBallRotate;    // Rotation component
static Matrix mBallInvRotate; // Inverse rotation
static Matrix mBallScale;
static Matrix mBallTranslate;

static Matrix mBallCamera;
static Vector vBallNormMouseLast;
static FLOAT  fBallScale = 1.0;

// track ball
static int iWidgetLastX = 0;
static int iWidgetLastY = 0;
static int centerx = 0;       // keep track of the sphere center
static int centery = 0;       
static int mousex, mousey;    // last x, y in viewport coords
static float curquat[4];      // current rotation quaternion 
static float lastquat[4];     // last rotation quaternion 
static float invlastquat[4];  // inverse of last rotation quaternion 
static int spinning;          // spin sphere
static int time;              // time between mouse motions 
static float velocity;        // speed of current mouse motion

//////////////////////////////////////////////////////////////////////////////////
// Functions

VOID DrawSphere(FLOAT fRadius, UINT_32 uiZSlices, UINT_32 uiThetaSlices )
{
  UINT_32 uiZSlice, uiThetaSlice;
  FLOAT fZMin, fZMax, fPhiMin, fPhiMax, fThetaMin, fThetaMax;
  FLOAT fPhiDelta, fThetaDelta;
  FLOAT fPhi0, fPhi1, fZ0, fZ1, fR0, fR1, fTheta;
  Vector vNormal;
  Point pppt[2][2];

  fZMin = -1;
  fZMax =  1;
  fPhiMin = DEGTORAD(-90);
  fPhiMax = DEGTORAD(90);
  fThetaMin = 0.0;
  fThetaMax = DEGTORAD(360);
  
  fPhiDelta = (fPhiMax - fPhiMin) / (FLOAT)uiZSlices;
  fThetaDelta = (fThetaMax - fThetaMin) / (FLOAT)uiThetaSlices;
  
  fPhi0 = fPhiMin;
  fPhi1 = fPhi0 + fPhiDelta;
  for ( uiZSlice = 0; uiZSlice < uiZSlices; uiZSlice++ )
    {
      fZ1 = fRadius*(FLOAT)sin(fPhi1);
      fR1 = fRadius*(FLOAT)cos(fPhi1);
      fZ0 = fRadius*(FLOAT)sin(fPhi0);
      fR0 = fRadius*(FLOAT)cos(fPhi0);
      
      glBegin(GL_TRIANGLE_STRIP);
      {
	fTheta = fThetaMin;
	  
	// Must seed the for loop with the left most two points
	pppt[0][0].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	pppt[1][0].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	pppt[1][1].Set(fR1*(FLOAT)cos(fTheta+fThetaDelta), 
		       fR1*(FLOAT)sin(fTheta+fThetaDelta), fZ1);
	  
	vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	vNormal.Normalize();
	  
	glNormal3fv(vNormal.Get());
	glVertex3fv(pppt[1][0].Get());
	glVertex3fv(pppt[0][0].Get());
	
	fTheta += fThetaDelta;	
	for ( uiThetaSlice = 0; uiThetaSlice < uiThetaSlices; uiThetaSlice++ )
	  {
	    pppt[1][1].Set(fR1*(FLOAT)cos(fTheta), fR1*(FLOAT)sin(fTheta), fZ1);
	    pppt[0][1].Set(fR0*(FLOAT)cos(fTheta), fR0*(FLOAT)sin(fTheta), fZ0);
	    
	    vNormal = (pppt[0][0] - pppt[1][0]) * (pppt[1][1] - pppt[0][0]);
	    vNormal.Normalize();
	    glNormal3fv(vNormal.Get());
	    glVertex3fv(pppt[1][1].Get());
	    
	    vNormal = (pppt[1][1] - pppt[0][1]) * (pppt[0][0] - pppt[1][1]);
	    vNormal.Normalize();
	    glNormal3fv(vNormal.Get());
	    glVertex3fv(pppt[0][1].Get());
	      
	    pppt[0][0] = pppt[0][1];
	    pppt[1][0] = pppt[1][1];
	    fTheta += fThetaDelta;
	  }
      }
      glEnd(); 

      fPhi0 = fPhi1;
      fPhi1 += fPhiDelta;
    }
}

VOID Init()
{
  GLuiSphere1_30  = glGenLists(1);
  GLuiSphere01_10 = glGenLists(1);
  GLuiSphere02_10 = glGenLists(1);

  glNewList( GLuiSphere1_30, GL_COMPILE );
  {
    DrawSphere(1, 30, 30);
  }
  glEndList();

  glNewList( GLuiSphere01_10, GL_COMPILE );
  {
    DrawSphere(0.01, 10, 10);
  }
  glEndList();

  glNewList( GLuiSphere02_10, GL_COMPILE );
  {
    DrawSphere(0.02, 10, 10);
  }
  glEndList();
}

VOID SetStatusLabel(CHAR *pcLabel)
{
  int iCode;

  sprintf(gpcBuffer, "SetStatusLabel \"%s\"", pcLabel);
  iCode = Tcl_Eval(pTclInterp, gpcBuffer);

  if ( iCode != TCL_OK )
    {
      printf( "Error: %s:\n", gpcBuffer );
      printf("Error Message: %s \n", pTclInterp->result);
    }
}

FLOAT Norm2ModelSpace(FLOAT fNorm)
{
  return (5.0/2.0)*fNorm/fBallScale;
}

BOOL SphereIntersection( Point &pt, struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY)
{
  FLOAT f, fX, fY, fNormX, fNormY, fNormCenterX, fNormCenterY;
  INT_32 iWidth, iHeight;

  iWidth = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  // Normalize widget coords
  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;
  fNormCenterX = (float)(centerx + 250) / iWidth ;
  fNormCenterY = (float)(iHeight - (centery + 250)) / iHeight;

  // Calculate x and y in viewport -> model space distances
  fX = Norm2ModelSpace( fNormX-fNormCenterX );
  fY = Norm2ModelSpace( fNormY-fNormCenterY );
  pt.Set( fX, fY, 0.0 );

  // Check if the cursor is inside the circle
  f = 1.0 - pt[X]*pt[X] - pt[Y]*pt[Y];
  if( f >= 0 )
    {
      pt[Z] = sqrt(f);
      pt[W] = 1.0;
      pt = mBallInvRotate * pt;
      return TRUE;
    }

  return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////
// Tcl Callbacks

VOID ResetCamera()
{
  mBallRotate.SetIdentity();
  mBallInvRotate.SetIdentity();
  mBallScale.SetIdentity();
  mBallTranslate.SetIdentity();
  
  mBallCamera.SetIdentity();
  fBallScale = 1.0;
  
  centerx = 0;
  centery = 0;
}

int CBHandleGUICreated( ClientData clientData, Tcl_Interp *pInterp, 
			int argc, char **argv)
{
  ASSERT( argc == 1 );

  bGUICreated = TRUE;

  return TCL_OK;
}

int CBHandleInnerPoints( ClientData clientData, Tcl_Interp *pInterp, 
			 int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( bDrawInnerPoints )
    {
      bDrawInnerPoints = FALSE;

      if( tptLastInnerPoint != NULL )
	{
	  tptLastInnerPoint->SetConnected(FALSE);
	}
    }
  else
    {
      bDrawInnerPoints = TRUE;
    }

  return TCL_OK;
}

int CBHandleSetSymmetryGroup( ClientData clientData, Tcl_Interp *pInterp, 
			      int argc, char **argv)
{
  ASSERT( argc == 3 );

  int iSymmetryGroup, iSymmetryGroupTile, iCode;

  iSymmetryGroup     = atoi( argv[1] );
  iSymmetryGroupTile = atoi( argv[2] );

  if( pSG != NULL )
    {
      delete pSG;
      pSG = NULL;
    }

  if( bGUICreated )
    {
      // Reset Interface
      bDrawInnerPoints = FALSE;
      sprintf(gpcBuffer, "resetInnerPoints");
      iCode = Tcl_Eval(pTclInterp, gpcBuffer);
      if ( iCode != TCL_OK )
	{
	  printf( "Error: %s:\n", gpcBuffer );
	  printf("Error Message: %s \n", pTclInterp->result);
	}

      if( (dtDrawType != dtSPHERE) && (dtDrawType != dtPOLYHEDRA) )
	{
	  dtDrawType = dtSPHERE;

	  sprintf(gpcBuffer, "set draw 0");
	  iCode = Tcl_Eval(pTclInterp, gpcBuffer);
	  if ( iCode != TCL_OK )
	    {
	      printf( "Error: %s:\n", gpcBuffer );
	      printf("Error Message: %s \n", pTclInterp->result);
	    }

	  ResetCamera();
	}
    }

  switch( iSymmetryGroup )
    {
    case sgtOTetrahedron:
      {
	// OTetrahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OTTile1(); 
	    }
	  break;
	  case(2):
	    {
	      pSG = new OTTile2(); 
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtSTetrahedron:
      {
	// STetrahedron
	pSG = new STetrahedron(); 
      }
    break;
    case sgtOOctahedron:
      {
	// OOctahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OOTile1(); 
	    }
	  break;
	  case(2):
	    {
	      pSG = new OOTile2(); 
	    }
	  break;
	  case(3):
	    {
	      pSG = new OOTile3(); 
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtOCube:
      {
	// OCube
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OCTile1(); 
	    }
	  break;
	  case(2):
	    {
	      pSG = new OCTile2(); 
	    }
	  break;
	  case(3):
	    {
	      pSG = new OCTile3(); 
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtOIcosahedron:
      {
	// OCube
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new OITile1(); 
	    }
	  break;
	  case(2):
	    {
	      pSG = new OITile2(); 
	    }
	  break;
	  case(3):
	    {
	      pSG = new OITile3(); 
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtODodecahedron:
      {
	// ODodecahedron
	switch(iSymmetryGroupTile)
	  {
	  case(1):
	    {
	      pSG = new ODTile1(); 
	    }
	  break;
	  case(2):
	    {
	      pSG = new ODTile2(); 
	    }
	  break;
	  case(3):
	    {
	      pSG = new ODTile3(); 
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }	
	  break;    
	  }
      }
    break;
    case sgtPentaDodecahedron:
      {
	// PentaDodecahedron
	pSG = new PentaDodecahedron(); 
      }
    break;
    default:
      {
	pSG = NULL;
	ASSERT( FALSE );      
      }
    break;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleSetDraw( ClientData clientData, Tcl_Interp *pInterp, 
		     int argc, char **argv)
{
  ASSERT( argc == 2 );

  dtDrawType = (DrawType)(atoi( argv[1] ));

  if((dtDrawType == dtSTEREO) || (dtDrawType == dtTRI_TILE) || (dtDrawType == dtSOLID))
    {
      pSG->TriangulateTile();
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleSetThickness( ClientData clientData, Tcl_Interp *pInterp, 
			  int argc, char **argv)
{
  ASSERT( argc == 2 );

  FLOAT fThickness;
  
  fThickness = atof( argv[1] );
  pSG->SetThickness(1.0 - fThickness);

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleDrawGrid( ClientData clientData, Tcl_Interp *pInterp, 
		      int argc, char **argv)
{
  ASSERT( argc == 2 );

  gtDrawGrid = (GridType)(atoi( argv[1] ));  
  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleTriangulate( ClientData clientData, Tcl_Interp *pInterp, 
			 int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      pSG->TriangulateTile();
      Togl_PostRedisplay( toglBall );
    }

  return TCL_OK;
}

int CBHandleOutputSTL( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .stl -filetypes {{\"STL Files\" {.stl}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->TriangulateTile();
      pSG->OutputSTL(fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

int CBHandleOutputEB( ClientData clientData, Tcl_Interp *pInterp, 
		       int argc, char **argv)
{
  ASSERT( argc == 1 );

  if( pSG != NULL )
    {
      INT_32 iCode;
      UINT_32 uiLength;
      FILE *fpOutputFile;
      CHAR *pcOutputFile;

      sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .eb -filetypes {{\"EB Files\" {.eb}}}");
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      
      uiLength = strlen(Tcl_GetStringResult(pInterp));
      if ( uiLength == 0 )
	{
	  // The user hit Cancel
	  return TCL_OK;
	}
  
      pcOutputFile = new CHAR[strlen(Tcl_GetStringResult(pInterp)) + 1];
      ASSERT( pcOutputFile != NULL );
      strcpy(pcOutputFile, Tcl_GetStringResult(pInterp));
      fpOutputFile = NULL;

      while ( fpOutputFile == NULL )
	{
	  fpOutputFile = fopen(pcOutputFile, "w");
	  if ( fpOutputFile == NULL )
	    {
	      sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", pcOutputFile);
	      iCode = Tcl_Eval(pInterp, gpcBuffer); 
	      ASSERT( iCode == TCL_OK );
	      
	      if ( strcmp(Tcl_GetStringResult(pInterp), "cancel") == 0 )
		{
		  // Cancel
		  goto cancel;
		}
	    }
	}

      pSG->TriangulateTile();
      pSG->OutputEB(fpOutputFile);

    cancel:
      if ( pcOutputFile != NULL )
	{
	  delete []pcOutputFile;
	}
      
      if ( fpOutputFile != stdout && fpOutputFile != NULL )
	{
	  ASSERT( fpOutputFile != NULL );
	  fclose(fpOutputFile);
	}
      
      fpOutputFile = NULL;
    }

  Togl_PostRedisplay( toglBall );

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// Togl Callbacks

int CBHandleCamera( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 3 );

  INT_32 iAction;
  iAction = atoi( argv[2] );

  switch( iAction )
    {
    case 1: // Reset
      {
	ResetCamera();
      }
    break;
    default:
      {
	ASSERT( FALSE );      
      }
    break;
    }
  
  Togl_PostRedisplay( togl );

  return TCL_OK;
}

void animate()
{
  /*
  if( spinning ) 
    {
      add_quats(lastquat, curquat, curquat);

      float m[4][4];
      float mi[4][4];
      MATRIX4D mm;
      Matrix mBall;
      
      // rotation
      build_rotmatrix( m, lastquat );

      mm[0]  = m[0][0];
      mm[1]  = m[0][1];
      mm[2]  = m[0][2];
      mm[3]  = m[0][3];
      mm[4]  = m[1][0];
      mm[5]  = m[1][1];
      mm[6]  = m[1][2];
      mm[7]  = m[1][3];
      mm[8]  = m[2][0];
      mm[9]  = m[2][1];
      mm[10] = m[2][2];
      mm[11] = m[2][3];
      mm[12] = m[3][0];
      mm[13] = m[3][1];
      mm[14] = m[3][2];
      mm[15] = m[3][3];
      
      mBall.Set(mm);
      mBallRotate = mBall * mBallRotate;
      mBallCamera = mBallTranslate * mBallRotate * mBallScale;

      // inverse
      build_rotmatrix( mi, invlastquat );

      mm[0]  = mi[0][0];
      mm[1]  = mi[0][1];
      mm[2]  = mi[0][2];
      mm[3]  = mi[0][3];
      mm[4]  = mi[1][0];
      mm[5]  = mi[1][1];
      mm[6]  = mi[1][2];
      mm[7]  = mi[1][3];
      mm[8]  = mi[2][0];
      mm[9]  = mi[2][1];
      mm[10] = mi[2][2];
      mm[11] = mi[2][3];
      mm[12] = mi[3][0];
      mm[13] = mi[3][1];
      mm[14] = mi[3][2];
      mm[15] = mi[3][3];
      
      mBall.Set(mm);
      mBallInvRotate = mBallInvRotate * mBall;
    }
  
  Togl_PostRedisplay( toglBall );
  glFlush();

  if(!spinning) 
    {
      Tcl_CancelIdleCall(NULL, NULL);
    }
  else
    {
      Tcl_DoWhenIdle((Tcl_IdleProc *) animate, NULL);
    }
    */
}

VOID PickPoint(struct Togl *togl, int iWidgetX, int iWidgetY)
{
  if( dtDrawType != dtSPHERE )
    return;

  GLuint selectBuf[500];
  GLint hits;
  GLint viewport[4];
  unsigned int i, j, c;
  GLuint names;
  FLOAT fZMin, f1, f2;

  //////////////////////////////////////////////////
  // Draw to selection buffer

  glGetIntegerv(GL_VIEWPORT, viewport);
  glSelectBuffer(500, selectBuf);
  (void) glRenderMode(GL_SELECT);

  glInitNames();
  glPushName(SELECTED_NOTHING); // push on dummy name

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  {
    glLoadIdentity();

    gluPickMatrix((GLdouble)iWidgetX, (GLdouble)(viewport[3]-iWidgetY), 5.0, 5.0, viewport);
    glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
    glScalef( fBallScale, fBallScale, 1 );

    // Camera
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    {
      glLoadIdentity();	               
      glMultMatrixf(mBallCamera.Get());  // Apply the Crystal Ball Matrix


      // This Render order is important!!!

      // Render the sphere so point won't pick points on back of sphere
      glCallList(GLuiSphere1_30);

      // Render and push names
      glPushName(SELECTED_OUTLINE_EDGE);
      {
	pSG->RenderSelectionSphereTileOutlineEdges();
      }
      glPopName();

      glPushName(SELECTED_INNER_EDGE);
      {
	pSG->RenderSelectionSphereTileInnerEdges();
      }
      glPopName();

      glPushName(SELECTED_OUTLINE_POINT);
      {
	pSG->RenderSelectionSphereTileOutlinePoints();
      }
      glPopName();

      glPushName(SELECTED_INNER_POINT);
      {
	pSG->RenderSelectionSphereTileInnerPoints();
      }
      glPopName();
    }
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
  }
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glFlush();

  hits = glRenderMode(GL_RENDER);

  // Draw to selection buffer
  //////////////////////////////////////////////////

  //////////////////////////////////////////////////
  // Process Hits

  //printf( "\nhits = %d\n", hits );

  uiSelectedType  = SELECTED_NOTHING;
  uiSelectedIndex = SELECTED_NOTHING;

  // Find which object was selected
  fZMin = INFINITY;
  for( i = 0, c = 0; i < (UINT_32)hits; i++ )
    {
      names = selectBuf[c];
      //printf(" names = %d: ", names);

      if( names > 10000 ) // this is some error, I'm not sure what
	{
	  printf("ERROR: selection names = %d: ", names);
	  return;
	}

      f1 = (float)selectBuf[c+1]/0x7fffffff;
      f2 = (float)selectBuf[c+2]/0x7fffffff;

      // Find nearest object (but take edges over sphere)
      if( (f1 < fZMin) ||
	  ((selectBuf[c+4] == SELECTED_OUTLINE_EDGE) && (f1 < fZMin + 0.01)) ||
	  ((selectBuf[c+4] == SELECTED_INNER_EDGE)   && (f1 < fZMin + 0.01)) )
	{
	  fZMin = f1;
	  	  
	  if( names == 3 )
	    {
	      uiSelectedType  = selectBuf[c+4];
	      uiSelectedIndex = selectBuf[c+5];
	    }
	}
      c += 3;

      for( j = 0; j < names; j++, c++ )
	{
	  //printf("%d ", selectBuf[c]);
	}

      //printf( "    zmin: %f  (%f)", f1, f2 );
      //printf( "\n" );
    }
  // Process Hits
  //////////////////////////////////////////////////

  //printf( "Selected: %lu %lu\n", uiSelectedType, uiSelectedIndex );
}

VOID MoveSelectedPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  ASSERT( (uiSelectedType == SELECTED_OUTLINE_POINT) ||
	  (uiSelectedType == SELECTED_INNER_POINT) );

  if( uiSelectedIndex == SELECTED_NOTHING )
    return;

  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      if( uiSelectedType == SELECTED_OUTLINE_POINT)
	{
	  pSG->MoveOutlinePoint(uiSelectedIndex, pt);
	}
      else
	{
	  pSG->MoveInnerPoint(uiSelectedIndex, pt);
	}
    }
}

VOID InsertNewPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  if( uiSelectedIndex == SELECTED_NOTHING )
    return;

  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      if( uiSelectedType == SELECTED_OUTLINE_EDGE )
	{
	  pSG->InsertOutlinePoint(uiSelectedIndex, pt);
	  PickPoint( togl, iWidgetX, iWidgetY );
	}
      else if( uiSelectedType == SELECTED_INNER_EDGE )
	{
	  pSG->InsertInnerPoint(uiSelectedIndex, pt);
	  PickPoint( togl, iWidgetX, iWidgetY );
	}
    }
}

VOID AddNewInnerPoint( struct Togl *togl, INT_32 iWidgetX, INT_32 iWidgetY )
{
  Point pt;

  if( SphereIntersection( pt, togl, iWidgetX, iWidgetY ) )
    {
      tptLastInnerPoint = pSG->AddInnerPoint(pt);
      PickPoint( togl, iWidgetX, iWidgetY );
    }
}

VOID DeletePoint( struct Togl *togl )
{
  bDeletePoint = TRUE;

  if( uiSelectedType == SELECTED_OUTLINE_POINT )
    {
      pSG->DeleteOutlinePoint(uiSelectedIndex);
    }
  else if (uiSelectedType == SELECTED_INNER_POINT )
    {
      pSG->DeleteInnerPoint(uiSelectedIndex);
    }
  
  uiSelectedType  = SELECTED_NOTHING;
  uiSelectedIndex = SELECTED_NOTHING;
}

int CBHandleKeyboard( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 6 );

  INT_32 iKey, iState, iWidgetX, iWidgetY;

  iKey     = (int) argv[2][0];
  iState   = atoi( argv[3] );
  iWidgetX = atoi( argv[4] );
  iWidgetY = atoi( argv[5] );

  switch( iKey )
    {
    case 'd':
      {
	if( uiSelectedType == SELECTED_OUTLINE_POINT )
	  {
	    DeletePoint( togl );
	    Togl_PostRedisplay( togl );
	  }
      }
    break;
    default:
      {
      }
    break;    
    }

  return TCL_OK;
}

int CBHandleMouseButton( struct Togl *togl, int argc, char *argv[] )
{
  ASSERT( argc == 7 );

  SetStatusLabel( " " );

  INT_32 iButton, iPress, iWidgetX, iWidgetY, iTime, iWidth, iHeight;
  FLOAT fNormX, fNormY;
  Vector v;
 
  iButton  = atoi( argv[2] );
  iPress   = atoi( argv[3] );
  iWidgetX = atoi( argv[4] );
  iWidgetY = atoi( argv[5] );
  iTime    = atoi( argv[6] );

  iWidth = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;

  if( iButton == 1 ) // Left Button
    {
      bLeftButtonDown = (iPress == 0);

      if( iPress == 0 ) // Button Press
	{
	  spinning = 0;	 

	  if( bDrawInnerPoints )
	    {
	      AddNewInnerPoint(togl, iWidgetX, iWidgetY);
	      Togl_PostRedisplay( togl );
	    }
	  else
	    {
	      PickPoint(togl, iWidgetX, iWidgetY);
	      
	      if( ( uiSelectedType == SELECTED_OUTLINE_EDGE ) ||
		  ( uiSelectedType == SELECTED_INNER_EDGE ) )
		{
		  InsertNewPoint(togl, iWidgetX, iWidgetY);
		  Togl_PostRedisplay( togl );
		}
	    }
	  
	  mousex = iWidgetX;
	  mousey = iWidgetY;
	}
      else // Button Release
	{
	  if( spinning )
	    {
	      Tcl_DoWhenIdle((Tcl_IdleProc *) animate, NULL);
	    }

	  bDeletePoint = FALSE;
	  uiSelectedType  = SELECTED_NOTHING;
	  uiSelectedIndex = SELECTED_NOTHING;
	}
    }
  else if( iButton == 2 ) // Middle Button
    {
      bMiddleButtonDown = (iPress == 0);

      if( bLeftButtonDown )
	{
	  if( (uiSelectedType == SELECTED_OUTLINE_POINT) ||
	      (uiSelectedType == SELECTED_INNER_POINT) )
	    {
	      DeletePoint( togl );
	      Togl_PostRedisplay( togl );
	    }
	}
    }
  else if( iButton == 3 ) // Right Button
    {
      bRightButtonDown = (iPress == 0);

      if( iPress == 0 ) // Button Press
	{
	  iWidgetLastX = iWidgetX;
	  iWidgetLastY = iWidgetY;
	}
    }

  time = iTime;
  vBallNormMouseLast.Set( fNormX, fNormY, 0.0 );
      
  return TCL_OK;
}

int CBHandleMouseMotion( struct Togl *togl, int argc, char *argv[] )
{
  if( bDeletePoint )
    {
      return TCL_OK;
    }

  ASSERT( argc == 6 );

  INT_32 iButton, iWidgetX, iWidgetY, iTime, iWidth, iHeight;
  Vector vCurr, vMouseLast;
  FLOAT fNormX, fNormY;
  Matrix mBall;

  iButton  = atoi( argv[2] );
  iWidgetX = atoi( argv[3] );
  iWidgetY = atoi( argv[4] );
  iTime    = atoi( argv[5] );

  iWidth  = Togl_Width( togl );
  iHeight = Togl_Height( togl );

  fNormX = (float)iWidgetX / iWidth;
  fNormY = (float)(iHeight - iWidgetY) / iHeight;

  vCurr.Set( fNormX, fNormY, 0.0 );
  vMouseLast = vBallNormMouseLast;

  switch( iButton )
    {
    case 1:
      {
	// Left

	if( uiSelectedType != SELECTED_NOTHING )
	  {
	    if( (uiSelectedType == SELECTED_OUTLINE_POINT) ||
		(uiSelectedType == SELECTED_INNER_POINT) )
	      {
		MoveSelectedPoint(togl, iWidgetX, iWidgetY);
	      }

	    break; // Don't rotate
	  }
	
	FLOAT m[4][4];
	FLOAT mi[4][4];
	FLOAT acceleration, fVelocity;
	MATRIX4D mm;
	int x, y;

	x = iWidgetX;
	y = iWidgetY;

	// speed = #pixels / millisecond
	fVelocity = abs(mousex - iWidgetX + mousey - iWidgetY) / (FLOAT)(iTime - time);
	acceleration = fVelocity - velocity / (FLOAT)(iTime - time);

	/*
	printf ("Delta Time: %d\n", iTime - time );
	printf ("Pixel's/Delta Time: %f\n",  
		abs(mousex - iWidgetX + mousey - iWidgetY) / (FLOAT)(iTime - time) );
		*/
	//printf ("Acceleration: %f\n",  acceleration );

	//if( fVelocity > 0.3)
	if( acceleration > 0.3)
	  {
	    spinning = 1;
	  }
	else
	  {
	    spinning = 0;
	  }

	/*
	printf( "(%f %f) (%f %f)\n", (2.0*mousex - 500) / 500, (500 - 2.0*mousey) / 500,
		(2.0*x - 500) / 500,      (500 - 2.0*y) / 500);
		*/
	
	trackball(lastquat, (2.0*(mousex-centerx)-500)/500, (500-2.0*(mousey-centery))/500,
		                 (2.0*(x-centerx)-500)/500,    (500 - 2.0*(y-centery))/500);

	trackball(invlastquat, (2.0*(x-centerx)-500)/500,    (500 - 2.0*(y-centery))/500,
		          (2.0*(mousex-centerx)-500)/500, (500-2.0*(mousey-centery))/500);
		                 
	build_rotmatrix( m, lastquat );
	build_rotmatrix( mi, invlastquat );

	// rotation
	mm[0] = m[0][0];
	mm[1] = m[0][1];
	mm[2] = m[0][2];
	mm[3] = m[0][3];
	mm[4] = m[1][0];
	mm[5] = m[1][1];
	mm[6] = m[1][2];
	mm[7] = m[1][3];
	mm[8] =  m[2][0];
	mm[9] =  m[2][1];
	mm[10] = m[2][2];
	mm[11] = m[2][3];
	mm[12] = m[3][0];
	mm[13] = m[3][1];
	mm[14] = m[3][2];
	mm[15] = m[3][3];

	mBall.Set(mm);
	mBallRotate = mBall * mBallRotate;

	// inverse
	mm[0]  = mi[0][0];
	mm[1]  = mi[0][1];
	mm[2]  = mi[0][2];
	mm[3]  = mi[0][3];
	mm[4]  = mi[1][0];
	mm[5]  = mi[1][1];
	mm[6]  = mi[1][2];
	mm[7]  = mi[1][3];
	mm[8]  = mi[2][0];
	mm[9]  = mi[2][1];
	mm[10] = mi[2][2];
	mm[11] = mi[2][3];
	mm[12] = mi[3][0];
	mm[13] = mi[3][1];
	mm[14] = mi[3][2];
	mm[15] = mi[3][3];

	mBall.Set(mm);
	mBallInvRotate = mBallInvRotate * mBall;

	mousex = iWidgetX;
	mousey = iWidgetY;
	velocity = fVelocity;
      }
    break;
    case 2:
      {
	// Middle

	/*
	if( bLeftButtonDown && (uiSelectedType == SELECTED_OUTLINE_POINT) )
	  {
	    break;
	  }
	  */

	
	/*
	FLOAT fDolly;

	fDolly = 2.0*((FLOAT)fNormY-vBallNormMouseLast[Y]);
	v.Set( 0.0, 0.0, fDolly);
	mBall.SetTranslate(v);
	*/

	FLOAT fScale, fNormCenterX, fNormCenterY;
	Point pt;
	Matrix m;

	fScale = pow(2, 5.0*((FLOAT)fNormY-vMouseLast[Y]));

	// Scaling the World
	//mBall.SetScale(fScale);
	//mBallScale = mBall * mBallScale;

	// Scaling the projection matrix (so light ok)
	fBallScale *= fScale;

	// Model center -> norm space -> widget space
	pt.Set( mBallTranslate.Get(0, 3), mBallTranslate.Get(1, 3), 0 );
	fNormCenterX = (2.0/5.0)*fBallScale * pt[X];
	fNormCenterY = (2.0/5.0)*fBallScale * pt[Y];	
	centerx = (int)( 500 * fNormCenterX);
	centery = (int)(-500 * fNormCenterY);
      }
    break;
    case 3:
      {
	// Right
	FLOAT fX, fY;
	Vector v;

	fX = Norm2ModelSpace( (FLOAT)fNormX-vBallNormMouseLast[X] );
	fY = Norm2ModelSpace( (FLOAT)fNormY-vBallNormMouseLast[Y] );

	v.Set( fX, fY, 0.0 );
	mBall.SetTranslate(v);

	mBallTranslate = mBall * mBallTranslate;
	centerx += (iWidgetX-iWidgetLastX);
	centery += (iWidgetY-iWidgetLastY);
      }
    break;
    default:
      {
	ASSERT( FALSE );      
      }
    break;
    }
  
  time = iTime;
  //mBallCamera = mBall * mBallCamera;
  mBallCamera = mBallTranslate * mBallRotate * mBallScale;
  vBallNormMouseLast.Set( fNormX, fNormY, 0.0 );
  iWidgetLastX = iWidgetX;
  iWidgetLastY = iWidgetY;


  Togl_PostRedisplay( togl );

  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////////////


// This is called by Tcl/Tk when the widget has been realized.  
// Here's where one may do some one-time context setup or initializations.
void CBHandleToglCreate( struct Togl *togl )
{
  toglBall = togl;
  
  glShadeModel(GL_FLAT);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_COLOR_MATERIAL);
  
  trackball(curquat, 0.0, 0.0, 0.0, 0.0);
}

// This is called by Tcl/Tk when the widget has been resized.  
// Typically, we call glViewport and perhaps setup the projection matrix.
void CBHandleToglReshape( struct Togl *togl )
{
  int width = Togl_Width( togl );
  int height = Togl_Height( togl );
  
  glViewport( 0, 0, width, height );
  
  // Set up projection transform
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
  
  // Change back to model view transform for rendering
  glMatrixMode(GL_MODELVIEW);
}

// This is called by Tcl/Tk when the widget's contents have to be redrawn.  
// Typically, we clear the color and depth buffers, render our objects, then 
// swap the front/back color buffers.
void CBHandleToglDisplay( struct Togl *togl )
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Lights
  FLOAT light_ambient[]  = { 0.2, 0.2, 0.2, 1.0 };
  FLOAT light_diffuse[]  = { 1.0, 1.0, 1.0, 1.0 };
  FLOAT light_specular[] = { 0.5, 0.5, 0.5, 1.0 };
  FLOAT position[] = { 5.0, 5.0, 5.0, 0.0 };   // Directional

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, position);

  // Camera & Scale
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1.25, 1.25, -1.25, 1.25, -100.0, 100.0);
  glScalef( fBallScale, fBallScale, 1 );
 
  // Model & Crystal Ball
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();	                 

  glEnable( GL_DEPTH_TEST );
  {
    /*
    // Draw Circle Center
    glPushMatrix();
    {
      FLOAT fX, fY;      
      fX = (float)centerx / 500;
      fY = (float)centery / 500;
      fX = Norm2ModelSpace( fX );
      fY = Norm2ModelSpace( fY );
      glColor3f(1, 0, 0);
      glTranslatef( fX, -fY, 1.05);
      glCallList(GLuiSphere02_10);
    }
    glPopMatrix();
    */

    glMultMatrixf(mBallCamera.Get());  // Apply the Crystal Ball Matrix

    if( pSG != NULL )
      {
	switch( dtDrawType )
	  {
	  case dtSPHERE:
	    {
	      pSG->RenderBall();	      

	      /*
		glCallList(GLuiSphere1_30);
	      pSG->RenderSelectionSphereTileOutlinePoints();
	      pSG->RenderSelectionSphereTileOutlineEdges();
	      pSG->RenderSelectionSphereTileInnerPoints();
	      pSG->RenderSelectionSphereTileInnerEdges();
	      */
	    }
	  break;
	  case dtPOLYHEDRA:
	    {
	      pSG->RenderPolyhedra();
	    }
	  break;
	  case dtSTEREO:
	    {
	      pSG->RenderStereographicPoints();
	    }
	  break;
	  case dtTRI_TILE:
	    {
	      pSG->RenderSphereTriangulation();
	    }
	  break;
	  case dtSOLID:
	    {
	      pSG->RenderSolid();
	    }
	  break;
	  default:
	    {
	      ASSERT(FALSE);
	    }
	  break;
	  }
      }
  }
  glDisable( GL_DEPTH_TEST );

  glMatrixMode(GL_MODELVIEW);  
  glLoadIdentity();  
  Togl_SwapBuffers( togl );
}

// Called by Tk_Main() to let me initialize the modules (Togl) I will need.
int Tcl_AppInit( Tcl_Interp *pInterp )
{
  int iCode;

  pTclInterp = pInterp;

  // Initialize Tcl, Tk, and the Togl widget module.
  if (Tcl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Tk_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
  if (Togl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

#ifdef WIN32
  // Set up a console window. Delete the following statement if you do not need that.
  if (TkConsoleInit(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
#endif // WIN32
  
  // Specify the C callback functions for widget creation, display, and reshape.
  Togl_CreateFunc(  CBHandleToglCreate  );
  Togl_DisplayFunc( CBHandleToglDisplay );
  Togl_ReshapeFunc( CBHandleToglReshape );

  // Make a new Togl widget command so the Tcl code can set a C variable.
  Togl_CreateCommand( "keyboard", CBHandleKeyboard );
  Togl_CreateCommand( "mouseButton", CBHandleMouseButton );
  Togl_CreateCommand( "mouseMotion", CBHandleMouseMotion );
  Togl_CreateCommand( "camera", CBHandleCamera );


  // Create Tcl Commands
  if( Tcl_CreateCommand(pInterp, "guiCreated", CBHandleGUICreated, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setSymmetryGroup", CBHandleSetSymmetryGroup, NULL, NULL) 
      == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setDraw", CBHandleSetDraw, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "setThickness", CBHandleSetThickness, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "drawGrid", CBHandleDrawGrid, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "innerPoints", CBHandleInnerPoints, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "triangulate", CBHandleTriangulate, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputSTL", CBHandleOutputSTL, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }
  if( Tcl_CreateCommand(pInterp, "outputEB", CBHandleOutputEB, NULL, NULL) == NULL )
    {
      return TCL_ERROR;
    }

  // Specify a user-specific startup file to invoke if the application is run interactively.
  iCode = Tcl_EvalFile(pInterp, "main.tcl"); 
  if ( iCode != TCL_OK )
    {
      sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error opening Main configure file: \\\"main.tcl\\\"\\n Error Message: %s\n Exiting program.\"", 
	      pInterp->result);
      iCode = Tcl_Eval(pInterp, gpcBuffer); 
      ASSERT( iCode == TCL_OK );
      return TCL_ERROR;
    }

  Init();
  
  return TCL_OK;
}

int main( int argc, char *argv[] )
{
  Tk_Main( argc, argv, Tcl_AppInit );
  return 0;
}

