#include "oitile3.h"

//////////////////////////////////////////////////////////////////////
// OITile3 Class
//

OITile3::OITile3()
{
  Vector v, v1, v2, v3, v4;
  FLOAT f, theta, y, z;
  Matrix m;
  Point pt;

  theta = atan( GOLDENRATIO / (GOLDENRATIO - 1.0) );
  y = sin(theta) / sqrt(3.0);
  z = GOLDENRATIO - (cos(theta) / sqrt(3.0));

  pt.Set(0.0, y, z);
  v.Set(0.0, GOLDENRATIO, 1.0);
  m.SetRotate(v, DEGTORAD(72));
  pt = m * pt;

  v1.Set(0.0, y, z);
  v2.Set(0.5, GOLDENRATIO/2.0, (1.0 + GOLDENRATIO)/2.0);
  v3.Set(pt[X], pt[Y], pt[Z] );
  v4.Set(0.0, GOLDENRATIO, 1.0);

  f = 1.0 / v4.Magnitude();

  v1 *= f;
  v2 *= f;
  v3 *= f;
  v4 *= f;

  // 2D
  m_pt2DTilePoint1.Set(v1[X], v1[Y], v1[Z]);
  m_pt2DTilePoint2.Set(v2[X], v2[Y], v2[Z]);
  m_pt2DTilePoint3.Set(v3[X], v3[Y], v3[Z]);
  m_pt2DTilePoint4.Set(v4[X], v4[Y], v4[Z]);

  Init();
}

OITile3::~OITile3()
{
}

VOID OITile3::InitOutline()
{
  TilePoint *tpt1, *tpt2, *tpt3, *tpt4;

  // 2D
  tpt1 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint1, tpTILE, 1 );
  tpt2 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint2, tpTILE, 2 );
  tpt3 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint3, tpTILE, 3 );
  tpt4 = m_pTile->InsertOutlinePoint(m_pt2DTilePoint4, tpTILE, 4 );

  tpt1->SetPartner(tpt3);
  tpt2->SetPartner(tpt2);
  tpt3->SetPartner(tpt1);
  tpt4->SetPartner(tpt4);

  m_pTile->TesselateOutlineAll();

  tpt1->SetHeightPartner(tpt3);
  tpt3->SetHeightPartner(tpt1);
}

///////////////////////////////////////////////////////////////////////////////

VOID OITile3::FindMatrices(Matrix &m1, Matrix &m2, TilePoint *tpt1)
{
  Vector v;

  if( tpt1->GetType() != tpTILE )
    {
      m1 = tpt1->GetPartnerMatrix();
      m2 = (tpt1->GetPartner())->GetPartnerMatrix();
    }
  else if( tpt1->GetTilePoint() == 1 )
    {
      v.Set(0.5, GOLDENRATIO/2.0, (1.0 + GOLDENRATIO)/2.0);
      m1.SetRotate(v, DEGTORAD(180));
      m2.SetRotate(v, DEGTORAD(-180));
    }
  else if( tpt1->GetTilePoint() == 2 )
    {
      v.Set(0.5, GOLDENRATIO/2.0, (1.0 + GOLDENRATIO)/2.0);
      m1.SetRotate(v, DEGTORAD(-180));
      m2.SetRotate(v, DEGTORAD(180));
    }
  else if( tpt1->GetTilePoint() == 3 )
    {
      v.Set(0.0, GOLDENRATIO, 1.0);
      m1.SetRotate(v, DEGTORAD(-72));
      m2.SetRotate(v, DEGTORAD(72));
    }
  else if( tpt1->GetTilePoint() == 4 )
    {
      v.Set(0.0, GOLDENRATIO, 1.0);
      m1.SetRotate(v, DEGTORAD(72));
      m2.SetRotate(v, DEGTORAD(-72));
    }
  else
    {
      ASSERT(FALSE);
    }
}

///////////////////////////////////////////////////////////////////////////////

VOID OITile3::OutputES(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "oicosahedron tile 3\n\n" );

  m_pTile->OutputES(fpOutputFile);

  OutputESColor(fpOutputFile);
}
