#include "tilepoint.h"
#include "snappoint.h"

#include <GL/gl.h>


//////////////////////////////////////////////////////////////////////
// TilePoint Class
//


TilePoint::TilePoint()
{
  m_uiID = 0;

  m_ptpPartner = NULL;
  m_ptpHeightPartner = NULL;

  m_uiTilePoint = 0;
  m_fHeight = 0;

  m_bConnected = TRUE;
  m_bHeightSelected = FALSE;

  m_pSnapPoint = NULL;

  m_listTesselatedPoint.Init();
}

TilePoint::~TilePoint()
{
}

VOID TilePoint::RemoveAllTesselatedPoints()
{
  m_listTesselatedPoint.RemoveAll();
}

VOID TilePoint::InsertTesselatedPoint(Point p)
{
  Point *pt;

  pt = new Point();
  pt->Set(p.Get());

  m_listTesselatedPoint.InsertLast(pt, 0);
}

VOID TilePoint::ClearSnapPoint()
{
  if( m_pSnapPoint != NULL )
    {
      TilePoint *tpt;

      tpt = m_pSnapPoint->RemoveTilePoint(this);

      if( tpt != NULL )
	{
	  tpt->DeleteSnapPoint();
	}

      m_pSnapPoint = NULL;
    }
}

VOID TilePoint::DeleteSnapPoint()
{
  if( m_pSnapPoint != NULL )
    {
      delete m_pSnapPoint;
      m_pSnapPoint = NULL;
    }
}

VOID TilePoint::CalculateHeightPoint()
{
  Vector v;

  v.Set( m_pt[X], m_pt[Y], m_pt[Z] );
  v.Normalize();
  v *= (1.0 + m_fHeight);

  m_ptHeight.Set(v[X], v[Y], v[Z]);
}

VOID TilePoint::AddHeight(FLOAT fHeight)
{
  m_fHeight += fHeight;
  CalculateHeightPoint();

  if( m_pSnapPoint != NULL )
    {
      m_pSnapPoint->SnapTilePoints(m_fHeight, m_ptHeight);
    }
}

VOID TilePoint::SnapToHeight(FLOAT fHeight, Point pt)
{
  m_fHeight = fHeight;
  m_ptHeight = pt;
}

VOID TilePoint::CalculateCentroid(Point &ptCentroid, UINT_32 &uiPoints)
{
  CListIterInt iter;
  Point *pt;
  Point p;

  iter.Init(&m_listTesselatedPoint);
  for( pt = (Point *)(iter.PeekFirst()); 
       pt != NULL;
       pt = (Point *)(iter.PeekNext()) )
    {
      p.Set(pt->Get());

      uiPoints++;
      ptCentroid += p;
    }
}

VOID TilePoint::PrintVertexTesselatedPoints()
{
  CListIterInt iter;
  Point *pt;

  iter.Init(&m_listTesselatedPoint);
  for( pt = (Point *)(iter.PeekFirst()); 
       pt != NULL;
       pt = (Point *)(iter.PeekNext()) )
    {
      glVertex3fv(pt->Get());
    }
}

VOID TilePoint::PrintVertexRaisedTesselatedPoints()
{
  CListIterInt iter;
  Point *pt;
  Point p;

  iter.Init(&m_listTesselatedPoint);
  for( pt = (Point *)(iter.PeekFirst()); 
       pt != NULL;
       pt = (Point *)(iter.PeekNext()) )
    {
      p.Set(pt->Get());
      p = 1.0025 * p;
      glVertex3fv(p.Get());
    }
}

////////////////////////////////////////////////////////////////////////////////

VOID TilePoint::OutputES(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "pt %f %f %f ", m_pt[X], m_pt[Y], m_pt[Z] );
  fprintf(fpOutputFile, "type %ld ", (UINT_32)m_tpType );

  switch( m_tpType )
    {
    case tpTILE:
      {
	fprintf(fpOutputFile, "tilepoint %ld ", (UINT_32)m_uiTilePoint );
      }
    break;
    case tpOUTLINE:
      {
	fprintf(fpOutputFile, "partnermatrix ");
	m_mPartnerTransform.OutputES(fpOutputFile);
      }
    break;
    case tpINNER:
      {
	fprintf(fpOutputFile, "connected %ld ", (UINT_32)m_bConnected );
	fprintf(fpOutputFile, "height %f ", m_fHeight );
      }
    case tpHOLE:
      {
	// nothing extra
      }
    break;
    default:
      {
	ASSERT(FALSE);
      }	
    break;  
    }
}

